import agentdojo.default_suites.v1_2.banking.injection_tasks
import agentdojo.default_suites.v1_2.slack.user_tasks
import agentdojo.default_suites.v1_2.travel.injection_tasks
import agentdojo.default_suites.v1_2.travel.user_tasks
import agentdojo.default_suites.v1_2.workspace.injection_tasks
import agentdojo.default_suites.v1_2.workspace.user_tasks  # noqa: F401
