import agentdojo.default_suites.v1_1_1.banking.user_tasks
import agentdojo.default_suites.v1_1_1.slack.user_tasks
import agentdojo.default_suites.v1_1_1.travel.injection_tasks
import agentdojo.default_suites.v1_1_1.travel.user_tasks
import agentdojo.default_suites.v1_1_1.workspace.user_tasks  # noqa: F401
